@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -28,6 +28,7 @@
 #include "gdevplnx.h"
 #include "gstrans.h"
 #include "gxdownscale.h"
+#include "gsbitops.h"
 
 #include "gdevkrnlsclass.h" /* 'standard' built in subclasses, currently First/Last Page and obejct filter */
 
@@ -210,6 +211,11 @@ gdev_prn_dev_spec_op(gx_device *pdev, int dev_spec_op, void *data, int size)
             return code;
     }
 
+#ifdef DEBUG
+    if (dev_spec_op == gxdso_debug_printer_check)
+        return 1;
+#endif
+
     return gx_default_dev_spec_op(pdev, dev_spec_op, data, size);
 }
 
@@ -227,6 +233,7 @@ gdev_prn_setup_as_command_list(gx_device *pdev, gs_memory_t *buffer_memory,
     gx_device_clist_common * const pcldev = &pclist_dev->common;
     bool reallocate = *the_memory != 0;
     byte *base;
+    bool save_is_open = pdev->is_open;	/* Save around temporary failure in open_c loop */
 
     while (target->parent != NULL) {
         target = target->parent;
@@ -297,8 +304,10 @@ open_c:
                                    "cmd list buf(retry open)");
             }
             ppdev->buf = *the_memory;
-            if (base != 0)
+            if (base != 0) {
+                pdev->is_open = save_is_open;	/* allow for success when we loop */
                 goto open_c;
+            }
         }
         /* Failure. */
         if (!reallocate) {
@@ -378,6 +387,7 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
     gs_memory_t *buffer_memory =
         (ppdev->buffer_memory == 0 ? pdev->memory->non_gc_memory :
          ppdev->buffer_memory);
+    bool deep = device_is_deep(pdev);
 
     /* If reallocate, find allocated memory & tear down buffer device */
     if (reallocate)
@@ -418,7 +428,7 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
             (&buf_space, pdev, NULL, pdev->height, false) >= 0;
         mem_space = buf_space.bits + buf_space.line_ptrs;
         if (ppdev->page_uses_transparency) {
-            pdf14_trans_buffer_size = (ESTIMATED_PDF14_ROW_SPACE(max(1, pdev->width), pdev->color_info.num_components) >> 3);
+            pdf14_trans_buffer_size = (ESTIMATED_PDF14_ROW_SPACE(max(1, pdev->width), pdev->color_info.num_components, deep ? 16 : 8) >> 3);
             if (new_height < (max_ulong - mem_space) / pdf14_trans_buffer_size) {
                 pdf14_trans_buffer_size *= pdev->height;
             } else {
@@ -1122,8 +1132,8 @@ gdev_prn_output_page_aux(gx_device * pdev, int num_copies, int flush, bool seeka
                 /* Print the accumulated page description. */
                 outcode = (*ppdev->printer_procs.print_page_copies)(ppdev, ppdev->file,
                                                           num_copies);
-                fflush(ppdev->file);
-                errcode = (ferror(ppdev->file) ? gs_note_error(gs_error_ioerror) : 0);
+                gp_fflush(ppdev->file);
+                errcode = (gp_ferror(ppdev->file) ? gs_note_error(gs_error_ioerror) : 0);
                 /* NB: background printing does this differently in its thread */
                 closecode = gdev_prn_close_printer(pdev);
 
@@ -1179,14 +1189,14 @@ gdev_prn_bg_output_page_seekable(gx_device * pdev, int num_copies, int flush)
 
 /* Print a single copy of a page by calling print_page_copies. */
 int
-gx_print_page_single_copy(gx_device_printer * pdev, FILE * prn_stream)
+gx_print_page_single_copy(gx_device_printer * pdev, gp_file * prn_stream)
 {
     return pdev->printer_procs.print_page_copies(pdev, prn_stream, 1);
 }
 
 /* Print multiple copies of a page by calling print_page multiple times. */
 int
-gx_default_print_page_copies(gx_device_printer * pdev, FILE * prn_stream,
+gx_default_print_page_copies(gx_device_printer * pdev, gp_file * prn_stream,
                              int num_copies)
 {
     int i = 1;
@@ -1203,9 +1213,9 @@ gx_default_print_page_copies(gx_device_printer * pdev, FILE * prn_stream,
          * right thing if we're producing multiple output files.
          * Code is mostly copied from gdev_prn_output_page.
          */
-        fflush(pdev->file);
+        gp_fflush(pdev->file);
         errcode =
-            (ferror(pdev->file) ? gs_note_error(gs_error_ioerror) : 0);
+            (gp_ferror(pdev->file) ? gs_note_error(gs_error_ioerror) : 0);
         closecode = gdev_prn_close_printer((gx_device *)pdev);
         pdev->PageCount++;
         code = (errcode < 0 ? errcode : closecode < 0 ? closecode :
@@ -1236,9 +1246,9 @@ prn_print_page_in_background(void *data)
 
     code = (*ppdev->printer_procs.print_page_copies)(ppdev, ppdev->file,
                                                           num_copies);
-    fflush(ppdev->file);
+    gp_fflush(ppdev->file);
 
-    errcode = (ferror(ppdev->file) ? gs_note_error(gs_error_ioerror) : 0);
+    errcode = (gp_ferror(ppdev->file) ? gs_note_error(gs_error_ioerror) : 0);
     bg_print->return_code = code < 0 ? code : errcode;
 
     /* Finally, release the foreground that may be waiting */
@@ -1321,8 +1331,8 @@ gdev_prn_open_printer_seekable(gx_device *pdev, bool binary_mode,
 
         if (seekable && !gp_fseekable(ppdev->file)) {
             errprintf(pdev->memory, "I/O Error: Output File \"%s\" must be seekable\n", ppdev->fname);
-            if (!IS_LIBCTX_STDOUT(pdev->memory, ppdev->file)
-              && !IS_LIBCTX_STDERR(pdev->memory ,ppdev->file)) {
+            if (!IS_LIBCTX_STDOUT(pdev->memory, gp_get_file(ppdev->file))
+              && !IS_LIBCTX_STDERR(pdev->memory, gp_get_file(ppdev->file))) {
 
                 code = gx_device_close_output_file(pdev, ppdev->fname, ppdev->file);
                 if (code < 0)
@@ -1466,17 +1476,22 @@ gx_default_create_buf_device(gx_device **pbdev, gx_device *target, int y,
         mdev = (gx_device_memory *)*pbdev;
     }
     if (target == (gx_device *)mdev) {
+        dev_t_proc_dev_spec_op((*orig_dso), gx_device) = dev_proc(mdev, dev_spec_op);
         /* The following is a special hack for setting up printer devices. */
         assign_dev_procs(mdev, mdproto);
+        /* Do not override the dev_spec_op! */
+        dev_proc(mdev, dev_spec_op) = orig_dso;
         check_device_separable((gx_device *)mdev);
         /* In order for saved-pages to work, we need to hook the dev_spec_op */
-        if (mdev->procs.dev_spec_op == NULL)
+        if (mdev->procs.dev_spec_op == NULL || mdev->procs.dev_spec_op == gx_default_dev_spec_op)
             set_dev_proc(mdev, dev_spec_op, gdev_prn_dev_spec_op);
 #ifdef DEBUG
         /* scanning sources didn't show anything, but if a device gets changed or added */
         /* that has its own dev_spec_op, it should call the gdev_prn_spec_op as well    */
-        else
-            errprintf(mdev->memory, "Warning: printer device has private dev_spec_op\n");
+        else {
+            if (dev_proc(mdev, dev_spec_op)((gx_device *)mdev, gxdso_debug_printer_check, NULL, 0) < 0)
+                errprintf(mdev->memory, "Warning: printer device has private dev_spec_op\n");
+        }
 #endif
         gx_device_fill_in_procs((gx_device *)mdev);
     } else {
