@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -20,11 +20,7 @@
 #  define gxdither_INCLUDED
 
 #include "gxfrac.h"
-
-#ifndef gx_device_halftone_DEFINED
-#  define gx_device_halftone_DEFINED
-typedef struct gx_device_halftone_s gx_device_halftone;
-#endif
+#include "gsdcolor.h"
 
 /*
  * Render DeviceN possibly by halftoning.
